# SpotRun

# ---------------------------- Globals ----------------
# Instruction location
InsCurrent = -1

# Number of steps so far
NumSteps = 1

# formatted values returned
ErrorMessage = ''
Output = ''

# Instructions, InputFile
Instructions = []
InputFile = []
Cards = {}

PushPopStack = []
GosubStack = []
LabelLines = {}

Specials = {}

Wordsize=0

import SpotWordsize

# --------------------------- SpotRun ----------------------------
# returns errormessage (already formatted) and output
def SpotExec(lstins, datafile, maxrun, wordsize, specials):
    global ErrorMessage, Output, Instructions, InputFile, NumSteps
    global Specials, Wordsize
    if len(specials)>0: Specials = specials
    Instructions = lstins
    InputFile = datafile
    Wordsize = wordsize
    
    # find the first instruction to execute
    FindNextInstruction()
    
    # gather labels into dictionary
    for i in range(len(Instructions)):
        ins = Instructions[i]
        if ins.type=='label': LabelLines[ins.label] = i

    # Main loop
    while True:
        command = lstins[InsCurrent].command
        if command=='next':
            if not ins_next(): break
        elif command=='read-prop':
            if not ins_read_prop(): break
        elif command=='rewind':
            if not ins_rewind(): break
        elif command=='say':
            if not ins_say('say'): break
        elif command=='saybin':
            if not ins_say('saybin'): break
        elif command=='sayhex':
            if not ins_say('sayhex'): break
        elif command=='write':
            if not ins_write(): break
        elif command=='write-from-ind':
            if not ins_write_from_ind(): break
        elif command=='write-to-ind':
            if not ins_write_to_ind(): break
        elif command=='push':
            if not ins_push(): break
        elif command=='pop':
            if not ins_pop(): break
        elif command=='stop':
            if not ins_stop(): break
        elif command=='jump':
            if not ins_goto(): break
        elif command[:4]=='jump':
            if not ins_jump(): break
        elif command=='gosub':
            if not ins_gosub(): break
        elif command=='return':
            if not ins_return(): break
        elif command=='inc':
            if not ins_inc(): break
        elif command=='dec':
            if not ins_dec(): break
        elif command=='add':
            if not ins_arith('add'): break
        elif command=='sub':
            if not ins_arith('sub'): break
        elif command=='mul':
            if not ins_arith('mul'): break
        elif command=='div':
            if not ins_arith('div'): break
        elif command=='mod':
            if not ins_arith('mod'): break
        elif command=='and':
            if not ins_arith('and'): break
        elif command=='or':
            if not ins_arith('or'): break
        elif command=='xor':
            if not ins_arith('xor'): break
        elif command=='rshift':
            if not ins_arith('rshift'): break
        elif command=='lshift':
            if not ins_arith('lshift'): break
        elif command=='comp':
            if not ins_comp(): break
        else:
            ErrorMessage = '<font color=red>Unknown command ('+command+') in line:</font><br>\n'+lstins[InsCurrent].sline
            break
        NumSteps += 1
        if NumSteps > maxrun:
            ErrorMessage = '<font color=red>Number of instructions steps exceeded '+str(maxrun)+'</font>'
            break
    return ErrorMessage, Output

# -------------------- FindNextInstruction() ----------------
# sets InsCurrent if possible.  If not, returns False, and sets ErrorMessage
def FindNextInstruction():
    global ErrorMessage, InsCurrent, NumSteps

    if InsCurrent < len(Instructions)-1:
        for i in range(InsCurrent+1, len(Instructions)):
            if Instructions[i].type == 'command':
                InsCurrent=i;
                return True
    
    ErrorMessage = '<font color=green>Program terminated after '+str(NumSteps)+' steps.</font>'+LastLine()
    return False    

# ------------------------- LastLine -------------------------
# returns string showing last line executed
def LastLine():
    s = '\n<br>Line #' + str(InsCurrent+1) + ' was last line executed:\n<br>'
    s += Instructions[InsCurrent].sline
    return s

# --------------------------- GetCardValue -------------------
def GetCardValue(cardno, MustBeInteger=False):
    if Cards.has_key(cardno):
        ovalue = Cards[cardno]
        sovalue = str(ovalue)
        if MustBeInteger and not (sovalue.isdigit() or (sovalue[0]=='-' and sovalue[1:].isdigit())):
            return '','<font color=red>Value of C'+str(cardno)+' is not an integer</font>'+LastLine()
        return ovalue,''
    return '','<font color=red>C' + str(cardno) + ') has no value</font>'+LastLine()

# ------------------------ ins_next() -------------------------
def ins_next():
    icurrent = InputFile['icurrent']
    if icurrent < InputFile['numrows']:
        InputFile['icurrent'] = icurrent + 1
    return FindNextInstruction()

# ---------------------------- ins_read_prop ------------------
def ins_read_prop():
    global Instructions, ErrorMessage, InputFile, Cards

    ins = Instructions[InsCurrent]
    icurrent = InputFile['icurrent']
    if icurrent >= InputFile['numrows']:
        ErrorMessage = '<font color=red>Trying to read past end-of-input-file<br>\n'
        ErrorMessage += 'Use JUMP-IF-EOF to prevent this.</font>'+LastLine()
        return False
    ifield = ins.args['arg0_value']
    if ifield < 1 or ifield > InputFile['numfields']:
        ErrorMessage = '<font color=red>Incorrect property number. (Properties for this input file range from 1 to '
        ErrorMessage += str(InputFile['numfields'])+')</font>'+LastLine()
        return False
    val = InputFile['rows'][icurrent][ifield-1]
    try:
        ival = int(val)
        Cards[ins.args['arg1_value']] = SpotWordsize.limitToWordsize(val,Wordsize)
    except:
        Cards[ins.args['arg1_value']] = val
    return FindNextInstruction()

# ---------------------------- ins_rewind ---------------------
def ins_rewind():
    InputFile['icurrent'] = 0
    return FindNextInstruction()

# ----------------------------- ins_say -----------------------
def ins_say(which):
    global Output, ErrorMessage
    
    ins=Instructions[InsCurrent]
    stype = ins.args['arg0_type']
    svalue = ins.args['arg0_value']
    if stype == 'number':
        if which=='say':
            Output += str(svalue)
        elif which=='saybin':
            Output += SpotWordsize.toBin(int(svalue),Wordsize)
        elif which=='sayhex':
            Output += SpotWordsize.toHex(int(svalue),Wordsize)
    elif stype == 'string':
        Output += svalue.replace('\\n','<br>\n')
    elif stype == 'card':
        if not Cards.has_key(svalue):
            ErrorMessage = '<font color=red>C'+str(svalue)+' has no value</font>'+LastLine()
            return False
        svalue = str(Cards[svalue])
        if (svalue.isdigit() or (svalue[0]=='-' and svalue[1:].isdigit())):
            ivalue = int(svalue)
            if which=='say':
                Output += svalue
            elif which=='saybin':
                Output += SpotWordsize.toBin(ivalue,Wordsize)
            elif which=='sayhex':
                Output += SpotWordsize.toHex(ivalue,Wordsize)
        else:
            Output += svalue.replace('\\n','<br>\n')
    elif stype == 'special':
        if Specials.has_key(svalue[1:].lower()):
            Output += Specials[svalue[1:].lower()]
        else:
            ErrorMessage = '<font color=red>"'+svalue+'" is not recognized.</font>'
            return False
        
    return FindNextInstruction()


# ------------------------------ ins_write -------------------
def ins_write():
    global ErrorMessage

    ins = Instructions[InsCurrent]
    type0 = ins.args['arg0_type']
    value0 = ins.args['arg0_value']
    if type0 == 'number':
        ovalue = SpotWordsize.limitToWordsize(value0,Wordsize)
    elif type0 == 'string':
        ovalue = value0
    elif type0 == 'card':
        ovalue,ErrorMessage = GetCardValue(value0)
        if ErrorMessage != '': return False

    Cards[ins.args['arg1_value']] = ovalue
        
    return FindNextInstruction()

# ---------------------------- ins_write_from_ind ---------------
def ins_write_from_ind():
    global ErrorMessage

    ins = Instructions[InsCurrent]
    type0 = ins.args['arg0_type']
    if type0 == 'number':
        add0 = int(ins.args['arg0_value'])
    else:
        add0,ErrorMessage = GetCardValue(ins.args['arg0_value'],True)
        if ErrorMessage != '': return False
        
    add1,ErrorMessage = GetCardValue(ins.args['arg1_value'],True)
    if ErrorMessage != '': return False
    
    add2 = add0 + add1
    if add2 < 0:
        ErrorMessage = '<font color=red>Base + Index results in '+str(add2)+' -- no such Card</font>'+LastLine()
        return False
    
    ovalue,ErrorMessage = GetCardValue(add2)
    if ErrorMessage != '': return False

    Cards[ins.args['arg2_value']] = ovalue
    return FindNextInstruction()

# ----------------------------- ins_write_to_ind ----------------
def ins_write_to_ind():
    global ErrorMessage

    ins=Instructions[InsCurrent]
    type1 = ins.args['arg1_type']
    if type1 == 'number':
        add1 = int(ins.args['arg1_value'])
    else:
        add1,ErrorMessage = GetCardValue(ins.args['arg1_value'],True)
        if ErrorMessage != '': return False

    add2,ErrorMessage = GetCardValue(ins.args['arg2_value'],True)
    if ErrorMessage != '': return False

    add3 = add1 + add2
    if add3 < 0:
        ErrorMessage = '<font color=red>Base + Index results in '+str(add3)+' -- no such Card</font>'+LastLine()
        return False

    ovalue,ErrorMessage = GetCardValue(ins.args['arg0_value'])
    if ErrorMessage != '': return False

    Cards[add3] = ovalue
        
    return FindNextInstruction()

# --------------------------- ins_push ----------------------
def ins_push():
    global ErrorMessage

    ins = Instructions[InsCurrent]
    ovalue,ErrorMessage = GetCardValue(ins.args['arg0_value'])
    if ErrorMessage != '' : return False
    PushPopStack.append(ovalue)
    
    return FindNextInstruction()

# ---------------------------- ins_pop -----------------------
def ins_pop():
    global ErrorMessage
    
    ins = Instructions[InsCurrent]
    if len(PushPopStack) == 0:
        ErrorMessage = '<font color=red>The Push stack is empty -- nothing to Pop</font>'+LastLine()
        return False
    Cards[ins.args['arg0_value']] = PushPopStack[-1]
    del PushPopStack[-1]
    
    return FindNextInstruction()

# ---------------------------- ins_stop ----------------------
def ins_stop():
    global ErrorMessage

    ErrorMessage='<font color=green>Normal program termination after '+str(NumSteps)+' steps.</font>'+LastLine()
    return False

# ------------------------------- ins_goto -------------------
def ins_goto():
    global ErrorMessage, InsCurrent

    ins = Instructions[InsCurrent]
    gotoplace = ins.args['arg0_value']
    if not LabelLines.has_key(gotoplace):
        ErrorMessage='<font color=red>Cannot find label: '+str(gotoplace)+'</font>'+LastLine()
        return False
    InsCurrent = LabelLines[gotoplace]
    return FindNextInstruction()

# -------------------------------- ins_jump -------------------
# Process all of the jumps
def ins_jump():
    global ErrorMessage, InsCurrent

    ins = Instructions[InsCurrent]
    command = ins.command
    gotoplace = ins.args['arg0_value']
    if not LabelLines.has_key(gotoplace):
        ErrorMessage='<font color=red>Cannot find label: '+str(gotoplace)+'</font>'+LastLine()
        return False

    if command[-3:] == 'eof':
        if InputFile['icurrent'] >= InputFile['numrows']:
            InsCurrent = LabelLines[gotoplace]
    else:
        val1,ErrorMessage = GetCardValue(ins.args['arg1_value'])
        if ErrorMessage != '': return False
        if ins.args['arg2_type'] == 'number' or ins.args['arg2_type'] == 'string':
            val2 = ins.args['arg2_value']
        elif ins.args['arg2_type'] == 'card':
            val2,ErrorMessage = GetCardValue(ins.args['arg2_value'])
            if ErrorMessage != '': return False

        jtype = command[-2:]
        dojump = False
        if jtype=='eq':
            if val1 == val2: dojump = True
        elif jtype=='ne':
            if val1 != val2: dojump = True
        elif jtype=='gt':
            if val1 > val2: dojump = True
        elif jtype=='ge':
            if val1 >= val2: dojump = True
        elif jtype=='le':
            if val1 <= val2: dojump = True
        elif jtype=='lt':
            if val1 < val2: dojump = True
        
        if dojump:
            InsCurrent = LabelLines[gotoplace]
            
    return FindNextInstruction()

# -------------------------------- ins_gosub ------------------
def ins_gosub():
    global ErrorMessage, InsCurrent

    ins = Instructions[InsCurrent]
    gosubplace = ins.args['arg0_value']
    if not LabelLines.has_key(gosubplace):
        ErrorMessage='<font color=red>Target ('+str(gosubplace)+') not found</font>'+LastLine()
        return False
    GosubStack.append(InsCurrent)
    InsCurrent = LabelLines[gosubplace]
    return FindNextInstruction()

# --------------------------------- ins_return ----------------
def ins_return():
    global ErrorMessage, InsCurrent

    if len(GosubStack) == 0:
        ErrorMessage = '<font color=red>No place to RETURN to.</font>'+LastLine()
        return False
    target = GosubStack[-1]
    del GosubStack[-1]
    if target >= len(Instructions):
        ErrorMessage = '<font color=red>RETURNing beyond the end of the program.</font>'+LastLine()
        return False
    InsCurrent = target
    return FindNextInstruction()

# -------------------------------- ins_inc ------------------
def ins_inc():
    global ErrorMessage

    ins = Instructions[InsCurrent]
    cardno = ins.args['arg0_value']
    ovalue,ErrorMessage = GetCardValue(cardno, True)
    if ErrorMessage != '': return False
    Cards[cardno] = SpotWordsize.limitToWordsize(ovalue + 1,Wordsize)
    return FindNextInstruction()

# -------------------------------- ins_dec -----------------
def ins_dec():
    global ErrorMessage

    ins = Instructions[InsCurrent]
    cardno = ins.args['arg0_value']
    ovalue,ErrorMessage = GetCardValue(cardno, True)
    if ErrorMessage != '': return False
    Cards[cardno] = SpotWordsize.limitToWordsize(ovalue - 1,Wordsize)
    return FindNextInstruction()

# ---------------------------------- ins_arith -------------------
def ins_arith(atype):
    global ErrorMessage

    ins = Instructions[InsCurrent]
    if ins.args['arg0_type'] == 'number':
        val1 = SpotWordsize.limitToWordsize(ins.args['arg0_value'],Wordsize)
    else:
        val1,ErrorMessage = GetCardValue(ins.args['arg0_value'],True)
        if ErrorMessage != '': return False
        
    if ins.args['arg1_type'] == 'number':
        val2 = SpotWordsize.limitToWordsize(ins.args['arg1_value'],Wordsize)
    else:
        val2,ErrorMessage = GetCardValue(ins.args['arg1_value'],True)
        if ErrorMessage != '': return False

    if atype == 'add':
        val3 = val1 + val2
    elif atype == 'sub':
        val3 = val1 - val2
    elif atype == 'mul':
        val3 = val1 * val2
    elif val2 == 0:
        ErrorMessage = '<font color=red>Division by zero.</font>'+LastLine()
        return False
    elif atype == 'div':
        val3 = val1 / val2
    elif atype == 'mod':
        val3 = val1 % val2
    elif atype == 'and':
        val3 = val1 & val2
    elif atype == 'or':
        val3 = val1 | val2
    elif atype == 'xor':
        val3 = val1 ^ val2
    elif atype == 'rshift':
        if val2 < 0:
            ErrorMessage = '<font color=red>Shift amount for RSHIFT must be >= 0</font>'+LastLine()
            return False
        val3 = val1 >> val2
    elif atype == 'lshift':
        if val2 < 0:
            ErrorMessage = '<font color=red>Shift amount for LSHIFT must be >= 0</font>'+LastLine()
            return False
        val3 = val1 << val2

    Cards[ins.args['arg2_value']] = SpotWordsize.limitToWordsize(val3,Wordsize)
    return FindNextInstruction()

# --------------------------------------- ins_comp -------------------------------------
def ins_comp():
    global ErrorMessage

    ins = Instructions[InsCurrent]
    val1,ErrorMessage = GetCardValue(ins.args['arg0_value'],True)
    if ErrorMessage != '': return False
    try:
        ival1 = int(val1)
        val2 = ~ival1
        Cards[ins.args['arg1_value']] = SpotWordsize.limitToWordsize(val2,Wordsize)
        return FindNextInstruction()
    except:
        ErrorMessage = '<font color=red>Complement must operate on a number</font>'+LastLine()
        return False
    
    
